
# clock_fun should be time.perf_counter, time.monotonic,
# timeit.default_timer (for best compatibility)

# start should be attached to win.callOnFlip()
# next_frame should be attached to win.callOnFlip()
from timeit import default_timer


class Timeline(object):
    """Keeps track of the amount of time between frames."""

    def __init__(self, clock_fun=default_timer):
        """Create the Timeline object.

        Parameters
        ----------
        clock_fun: function or method
            Used to get the current time. Defaults to timeit.default_timer,
            which picks the highest-resolution timer available given the platform
            and version of python.
        """
        self.clock_fun = clock_fun
        self.running = False

    def start(self):
        """Start taking the time."""
        self.running = True
        self.start_time = self.clock_fun()
        self.prev_frame_time = self.start_time
        self.prev_frame_dur = 0

    def stop(self):
        """Stop taking the time."""
        self.running = False
        self.start_time = self.clock_fun()
        self.prev_frame_time = self.start_time
        self.prev_frame_dur = 0

    def next_frame(self):
        """Record the time of the frame.

        If using psychopy, use with callOnFlip (so it's called immediately
        after the flip is done).
        """
        # should be run by win.callOnFlip
        now = self.clock_fun()
        if not self.running:
            return
        self.prev_frame_dur = now - self.prev_frame_time
        self.prev_frame_time = now

    @property
    def frame_time(self):
        """Time relative to the start time."""
        return self.prev_frame_time - self.start_time

    @property
    def delta_time(self):
        """Time between last two frames."""
        return self.prev_frame_dur



import ctypes
import struct
import serial
import numpy as np
from time import sleep
from toon.input.device import BaseDevice
from serial.tools import list_ports

# thanks to ROS http://docs.ros.org/fuerte/api/cyberglove/html/serial__glove_8hpp.html
# for commands
# http://web.cs.ucdavis.edu/~neff/papers/GloveCalibration_WangNeff_SCA2013.pdf
# for sensor layout
# https://static1.squarespace.com/static/559c381ee4b0ff7423b6b6a4/t/58ca1b16be6594e83fb42402/1489640216840/CyberGlove+III.pdf
# is mostly relevant, and also describes the responses following commands

thumb_fields = ['roll', 'mcp', 'pip']  # index "owns" abduction
finger_fields = ['mcp', 'pip', 'abd']  # index through pinky
palm_wrist = ['arch', 'pitch', 'yaw']


class ThumbData(ctypes.Structure):
    _fields_ = [(n, ctypes.c_double) for n in thumb_fields]


class FingerData(ctypes.Structure):
    _fields_ = [(n, ctypes.c_double) for n in finger_fields]


class WristData(ctypes.Structure):
    _fields_ = [(n, ctypes.c_double) for n in palm_wrist]


fingers = [(n, FingerData) for n in ['index', 'middle', 'ring', 'pinky']]


class GloveData(ctypes.Structure):
    _fields_ = [('thumb', ThumbData)]
    _fields_.extend(fingers)
    _fields_.append(('wrist', WristData))


class Cyberglove(BaseDevice):
    sampling_frequency = 150
    shape = (1,)
    ctype = GloveData

    def __init__(self, port=None, **kwargs):
        super(Cyberglove, self).__init__(**kwargs)
        self.port = port  # TODO: auto-detect using serial.tools.list_ports
        self.dev = None

    def enter(self):
        if not self.port:
            self.port = next(x.device for x in list_ports.comports() if x.product == 'USB-Serial Controller')
        self.dev = serial.Serial(self.port, 115200, timeout=0.02)
        self.dev.reset_input_buffer()
        sleep(0.1)
        # test whether the device is connected/on
        self.dev.write(b'l ?\r')
        # should echo back & give result
        if not self.dev.read():
            raise ValueError('Make sure the device is switched on.')
        self.dev.write(b'f 0\r')  # stop filtering
        self.dev.write(b't 1152 1\r')  # 100 Hz
        self.dev.write(b'u 0\r')  # don't transmit status
        self.dev.write(b'l 1\r')  # light on
        sleep(0.1)
        self.dev.reset_input_buffer()
        self.dev.write(b'S')  # start streaming
        self.dev.flush()
        # not sure how to get rid of the rest of the garbage,
        # so spin until we find the first b'S'
        for i in range(40):
            tmp = self.dev.read()
            if tmp == b'S':
                self.dev.read(19)  # read the rest of the line
                break
        if i >= 39:
            raise ValueError('Did not find the start byte.')

    def read(self):
        # one byte (S)
        val = self.dev.read(1)
        time = self.clock()
        if val:
            data = self.dev.read(19)  # read remaining bytes ('18 sensors + \x00')
            data = struct.unpack('<' + 'B' * 18, data[:-1])
            data = [(d - 1.0)/254.0 for d in data]
            # pack into proper location
            thumb_data = ThumbData(data[0], data[1], data[2])  # data[3] is abduction
            index_data = FingerData(data[4], data[5], data[3])
            middle_data = FingerData(data[6], data[7], data[8])
            ring_data = FingerData(data[9], data[10], data[11])
            pinky_data = FingerData(data[12], data[13], data[14])
            wrist_data = WristData(data[15], data[16], data[17])
            return time, GloveData(thumb=thumb_data,
                                   index=index_data, middle=middle_data,
                                   ring=ring_data, pinky=pinky_data,
                                   wrist=wrist_data)

    def exit(self):
        self.dev.write(b'\x03')  # stop streaming
        self.dev.write(b'l 0\r')  # light off
        sleep(0.1)
        self.dev.reset_input_buffer()
        self.dev.reset_output_buffer()
        self.dev.close()


if __name__ == '__main__':
    import time
    from toon.input.mpdevice import MpDevice
    dev = MpDevice(Cyberglove())
    with dev:
        start = time.time()
        while time.time() - start < 10:
            dat = dev.read()
            if dat is not None:
                print(dat)  # access joints via dat[-1]['thumb']['mcp']
            time.sleep(0.016)  # pretend to have a screen



from collections import namedtuple
from toon.anim.easing import linear
from toon.anim.interpolators import lerp, select


class Track(object):
    """Storage for (keyframe, value) pairs.

    Also handles interpolation and easing between keyframes.
    """

    def __init__(self, data, interpolator=lerp, easing=linear):
        """Creates a Track.

        Parameters
        ----------
        data: list of (keyframe, value) pairs. 
            These are expected to be sorted by keyframe in ascending order.
            Keyframe should not be duplicated.
        interpolator: function
            `lerp` for linear interpolation, `select` for stepwise behavior.
        easing: function
            Rate of change of the value over time. See toon.anim.easing, or
            specificy a custom function that takes a single parameter, and
            returns a value on the interval [0, 1].
        """
        # data is list of tuples
        self.data = data
        self.interpolator = interpolator
        self.easing = easing
        self.prev_index = 0
        # if data is non-numeric, force user to use select
        if not isinstance(data[0][1], (float, int)):
            self.interpolator = select
            self.easing = linear

    def at(self, time):
        """Get the interpolated value of a track at a given time.

        Parameters
        ----------
        time: float
            Time of interest.

        Returns
        -------
        The interpolated value at the given point in time.
        """

        # handle boundaries first
        data = self.data

        if time <= data[0][0]:
            self.prev_index = 0
            return data[0][1]

        if time >= data[-1][0]:
            self.prev_index = len(data) - 1
            return data[-1][1]

        # extract previous keyframe
        prev_index = self.prev_index
        prev_kf = data[prev_index]
        if time == prev_kf[0]:
            return prev_kf[1]

        sign = time - prev_kf[0]
        len_data = len(data)

        # if time is greater than previous keyframe time, search toward end
        if sign > 0:
            offset = -1
            for index in range(self.prev_index + 1, len_data):
                #print('a%i' % index)
                # find the next keyframe that the current time is less than
                proposed_kf = data[index]
                if time < proposed_kf[0]:
                    kf = proposed_kf
                    break
        # if time is less than previous keyframe time, search toward beginning
        elif sign < 0:
            offset = 1
            for index in range(self.prev_index - 1, -1, -1):
                #print('b%i' % index)
                # find the next keyframe that the current time is greater than
                proposed_kf = data[index]
                if time > proposed_kf[0]:
                    kf = proposed_kf
                    break
        else:
            raise ValueError('Unexpected value.')

        self.prev_index = index
        # find the other keyframe to interpolate between
        reference = data[index + offset]
        # print('kf_ref: %s, time: %s, kf_targ: %s' % (reference, time, kf))
        goal_time = kf[0] - reference[0]
        new_time = time - reference[0]
        time_warp = self.easing(1 - ((goal_time - new_time)/goal_time))
        return self.interpolator(reference[1], kf[1], time_warp)

    def duration(self):
        """The maximum duration of the track."""
        # last time in keyframes
        return self.data[-1][0]



from ctypes import c_double

import nidaqmx
import numpy as np
from nidaqmx.constants import AcquisitionType, TerminalConfiguration
from nidaqmx.stream_readers import AnalogMultiChannelReader
from nidaqmx._task_modules.read_functions import _read_analog_f_64
from toon.input.device import BaseDevice


class ForceKeyboard(BaseDevice):
    shape = (2,)
    ctype = c_double

    def __init__(self, sampling_frequency=250, indices=[7, 8], **kwargs):
        super(ForceKeyboard, self).__init__(**kwargs)
        self.sampling_frequency = sampling_frequency
        self._buffer = np.empty(self.shape, dtype=c_double)
        if len(indices) > 2:
            raise ValueError('Too many indices for ForceKeyboard.')
        self._indices = indices

    def enter(self):
        # assume first NI DAQ is the one we want
        self._device_name = nidaqmx.system.System.local().devices[0].name
        chans = [2, 9, 1, 8, 0, 10, 3, 11, 4, 12]
        sub_chans = [chans[i] for i in self._indices]
        channels = [self._device_name + ('/ai%i' % n) for n in sub_chans]
        channels = ','.join(channels)
        dev = nidaqmx.Task()
        dev.ai_channels.add_ai_voltage_chan(channels,
                                            terminal_config=TerminalConfiguration.RSE)
        dev.timing.cfg_samp_clk_timing(self.sampling_frequency,
                                       sample_mode=AcquisitionType.CONTINUOUS)
        self._reader = AnalogMultiChannelReader(dev.in_stream)
        dev.start()
        self._device = dev

    def read(self):
        #self._reader.read_one_sample(self._buffer, timeout=0.1)
        try:
            _read_analog_f_64(self._reader._handle, self._buffer, 1, 0)
        except Exception:
            return None
        # TODO: apply calibration?
        time = self.clock()
        return time, self._buffer

    def exit(self):
        self._device.stop()
        self._device.close()


if __name__ == '__main__':
    import time
    from toon.input import MpDevice
    dev = MpDevice(ForceKeyboard())

    times = []
    with dev:
        start = time.time()
        while time.time() - start < 240:
            dat = dev.read()
            if dat is not None:
                time, data = dat
                print(data)
                times.append(time)
            time.sleep(0.016)

    times = np.hstack(times)
    import matplotlib.pyplot as plt

    plt.plot(np.diff(times))
    plt.show()



from ctypes import c_double
import platform
import struct
import numpy as np

import hid
import usb.core
import usb.util
from toon.input.device import BaseDevice


def get_teensy_path(serial_number):
    hids = hid.enumerate()
    hids = [h for h in hids if h['product_id'] == 0x486 and h['vendor_id'] == 0x16c0]
    # if we have a serial number, filter on that too
    # otherwise, we'll end up taking the first
    if serial_number:
        hids = [h for h in hids if h['serial_number'] == serial_number]
    system = platform.system()
    if system == 'Darwin':
        hid_path = next(h['path'] for h in hids if h['usage'] == 512)
    elif system in ('Windows', 'Linux'):
        hid_path = next(h['path'] for h in hids if h['interface_number'] == 0)
    return hid_path


class Hand(BaseDevice):
    sampling_frequency = 1000
    ctype = c_double
    shape = (15,)

    def __init__(self, serial_number=None, blocking=True, **kwargs):
        super(Hand, self).__init__(**kwargs)
        self._inv_sqrt2 = 1/np.sqrt(2)
        self._device = None
        self._buffer = np.full(15, np.nan)
        self.serial_number = serial_number
        self.blocking = blocking

    def enter(self):
        self._device = hid.device()
        self._device.open_path(get_teensy_path(self.serial_number))
        self._device.set_nonblocking(not self.blocking)

    def exit(self):
        self._device.close()

    def read(self):
        data = self._device.read(46)
        time = self.clock()
        # timestamp, deviation from period, and 20x16-bit analog channels
        data = struct.unpack('>Lh' + 'H' * 20, bytearray(data))
        data = np.array(data, dtype='d')
        data[2:] /= 65535.0
        data[2:] -= 0.5
        self._buffer[0::3] = (data[2::4] - data[3::4]) * self._inv_sqrt2
        self._buffer[1::3] = (data[2::4] + data[3::4]) * self._inv_sqrt2
        self._buffer[2::3] = data[4::4] + data[5::4]
        return time, self._buffer



from timeit import default_timer as get_time


class MonoClock(object):
    """A stripped-down version of psychopy's clock.MonotonicClock.
    I wanted to avoid importing pyglet on the remote process, in case that causes any headache.
    """

    def __init__(self, relative=True):
        self._start_time = get_time() if relative else 0

    def get_time(self):
        """Returns the current time on this clock in seconds (sub-ms precision)
        """
        return get_time() - self._start_time

    def getTime(self):
        """Alias for get_time, so we can set the default psychopy clock in logging.
        """
        return self.get_time()

    @property
    def start_time(self):
        return self._start_time




import ctypes as c
from ctypes.util import find_library
from toon.input.device import BaseDevice

x11path = find_library('X11')
if not x11path:
    raise ImportError('Could not find the X11 library.')

x11 = c.cdll.LoadLibrary(x11path)
x11.XInitThreads()
# some tips:
# https://stackoverflow.com/questions/29638210/how-can-i-use-python-xlib-to-generate-a-single-keypress
# https://github.com/kripken/intensityengine/blob/master/src/python/intensity/components/thirdparty/Skype4Py/api/posix_x11.py
c_ulong_p = c.POINTER(c.c_ulong)
DisplayP = c.c_void_p
Atom = c.c_ulong
AtomP = c_ulong_p
XID = c.c_ulong
Window = XID
Bool = c.c_int
Status = c.c_int
Time = c.c_ulong
c_int_p = c.POINTER(c.c_int)

align = (c.sizeof(c.c_long) == 8 and c.sizeof(c.c_int) == 4)

_base_fields = [
    ('type', c.c_int),
    ('serial', c.c_ulong),
    ('send_event', Bool),
    ('display', DisplayP),
    ('window', Window),
    ('root', Window),
    ('subwindow', Window),
    ('time', Time),
    ('x', c.c_int),
    ('y', c.c_int),
    ('x_root', c.c_int),
    ('y_root', c.c_int),
    ('state', c.c_uint),
    # ('keycode', c.c_uint),
    ('same_screen', Bool),
]

key_fields = _base_fields.copy()
key_fields.insert(-1, ('keycode', c.c_uint))
button_fields = _base_fields.copy()
button_fields.insert(-1, ('button', c.c_uint))
motion_fields = _base_fields.copy()
motion_fields.insert(-1, ('is_hint', c.c_char))


class XKeyEvent(c.Structure):
    _fields_ = key_fields


class XButtonEvent(c.Structure):
    _fields_ = button_fields


class XMotionEvent(c.Structure):
    _fields_ = motion_fields


class XEvent(c.Union):
    _fields_ = [
        ('type', c.c_int),
        ('xkey', XKeyEvent),
        ('xbutton', XButtonEvent),
        ('xmotion', XMotionEvent),
        ('pad', c.c_long*24),
    ]


XEventP = c.POINTER(XEvent)
x11.XOpenDisplay.restype = DisplayP


class X11Mouse(BaseDevice):
    sampling_frequency = 1000  # optimistic
    shape = (1,)
    ctype = None

    def enter(self):
        disp = x11.XOpenDisplay(None)  # root window



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# get requirements
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()

# description for pypi
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    desc = f.read()

hand_reqs = ['hidapi', 'pyusb']
serial_reqs = ['pyserial']
force_reqs = ['nidaqmx;platform_system=="Windows"']
demo_reqs = ['matplotlib', 'pyqtgraph']
full_reqs = hand_reqs + serial_reqs + force_reqs + demo_reqs
full_reqs = list(set(full_reqs))

setup(
    name='toon',
    version='0.13.0',
    description='Tools for neuroscience experiments',
    long_description=desc,
    long_description_content_type='text/markdown',
    url='https://github.com/aforren1/toon',
    author='Alexander Forrence',
    author_email='aforren1@jhu.edu',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    install_requires=requirements,
    extras_require={
        'full': full_reqs,
        'demo': demo_reqs,
        'hand': hand_reqs,
        'birds': serial_reqs,
        'cyberglove': serial_reqs,
        'force': force_reqs
    },
    keywords='psychophysics neuroscience input experiment',
    packages=find_packages(exclude=['contrib', 'docs', 'tests'])
)



import struct
import time
from ctypes import c_double, Structure

import numpy as np
import serial
from serial.tools import list_ports

from toon.input.device import BaseDevice

# reference (most recent):
# https://github.com/aforren1/toon/blob/455d06827082ae30ec4ae3b2605185cb4d291c92/toon/input/birds.py

# with two birds, we can also try putting it in group mode & get adequate data rates


class Point3D(Structure):
    _fields_ = [('x', c_double), ('y', c_double), ('z', c_double)]


class BirdData(Structure):
    _fields_ = [('left', Point3D), ('right', Point3D)]


class Birds(BaseDevice):
    shape = (1,)
    ctype = BirdData

    def __init__(self, **kwargs):
        self._birds = None
        self._master = None
        self.indices = [1, 3]
        self.read_from = []
        self.cos_const = np.cos(-0.01938)
        self.sin_const = np.sin(0.01938)
        super(Birds, self).__init__(**kwargs)

    def enter(self):
        # timeout set so that we should always have data available
        devices = [x.device for x in list_ports.comports() if 'Keyspan' in x.description]
        self._birds = [serial.Serial(port, baudrate=115200, bytesize=serial.EIGHTBITS,
                                     xonxoff=0, rtscts=0, timeout=0.05) for port in devices]

        for bird in self._birds:
            bird.reset_input_buffer()
            bird.reset_output_buffer()
            bird.write(b'?')  # stop stream

        for bird in self._birds:
            bird.write(b'G')

        time.sleep(1)
        for bird in self._birds:
            bird.setRTS(0)

        # reorder birds
        out = [0 for i in range(len(self._birds))]
        for b in self._birds:
            b.write(b'\x4F' + b'\x15')
            res = b.read()
            res = struct.unpack('b', res)[0]
            out[res] = b
            if res == 1:
                self._master = b
        if self._master is None:
            raise ValueError('Master not found in ports provided.')
        self._birds = out  # overwrite unsorted
        # subset of birds to read
        for b in self._birds:
            if self.indices and res in self.indices:
                self.read_from.append(b)
        self.read_from.reverse()  # TODO: fix it up so that read_from is in order of bird indices
        # init master, FBB autoconfig
        time.sleep(1)
        self._master.write(('P' + chr(0x32) + chr(len(devices))).encode('utf-8'))
        time.sleep(3)

        # set the sampling frequency
        self._master.write(b'P' + b'\x07' + struct.pack('<H', int(130 * 256)))
        # check the sampling frequency
        # self._master.write(b'\x4F' + b'\x07')
        # time.sleep(0.1)
        # res = self._master.read(2)
        # print(struct.unpack('<H', res)[0]/256)

        for bird in self._birds:
            # change output type to position
            bird.write(b'V')
            # change Vm table to Ascension's "snappy" settings
            bird.write(b'P' + b'\x0C' + struct.pack('<HHHHHHH', *[2, 2, 2, 10, 10, 40, 200]))
            # first 5 bits are meaningless, B2 is 0 (AC narrow ON), B1 is 1 (AC wide OFF), B0 is 0 (DC ON)
            bird.write(b'P' + b'\x04' + b'\x02' + b'\x01')

        # ready to go, start streaming from birds whose indices were provided
        time.sleep(0.5)
        for b in self.read_from:
            b.write(b'@')

    def read(self):
        lst = []
        # busy wait until first byte available
        while not self.read_from[0].in_waiting:
            pass
        time = self.clock()
        for bird in self.read_from:
            lst.append(bird.read(6))  # assumes position data
        lst = [decode(msg) for msg in lst]
        data = np.array(lst).reshape((6,))  # position data for two birds
        data[:] = data[[1, 2, 0, 4, 5, 3]]  # fiddle with order of axes
        # rotate
        tmp_x = data[::3]
        tmp_y = data[1::3]
        data[::3] = tmp_x * self.cos_const - tmp_y * self.sin_const
        data[1::3] = tmp_y * self.sin_const + tmp_y * self.cos_const

        # translate to the lower left corner
        data[::3] += 61.35
        data[1::3] += 17.69
        out = BirdData(left=Point3D(*data[0:3]), right=Point3D(*data[3:6]))
        return time, out

    def exit(self):
        for bird in self.read_from:
            bird.write(b'?')  # stop stream
        time.sleep(1)
        self._master.write(b'G')  # sleep (master only?)
        time.sleep(1)
        for bird in self._birds:
            bird.close()


def decode(msg, n_words=3):
    return [decode_words(msg, i) for i in range(int(n_words))]


def decode_words(s, i):
    v = decode_word(s[2*i:2*i + 2])
    v *= 36 * 2.54  # scaling to cm
    return v / 32768.0


def decode_word(msg):
    lsb = msg[0] & 0x7f
    msb = msg[1]
    v = (msb << 9) | (lsb << 2)
    if v < 0x8000:
        return v
    return v - 0x10000


if __name__ == '__main__':
    import time
    from toon.input.mpdevice import MpDevice
    dev = MpDevice(Birds())
    times = []
    with dev:
        start = time.time()
        while time.time() - start < 20:
            dat = dev.read()
            if dat is not None:
                time, data = dat
                print(data['left'])
                times.append(np.diff(time))
            time.sleep(0.016)

    import matplotlib.pyplot as plt
    plt.plot(np.hstack(times))
    plt.show()
